package ca.recrown.islandsurvivalcraft.world.generation;

import org.bukkit.World;
import org.bukkit.block.Biome;

import ca.recrown.islandsurvivalcraft.Utilities;
import ca.recrown.islandsurvivalcraft.Types.Point2;
import ca.recrown.islandsurvivalcraft.caching.Cache;
import ca.recrown.islandsurvivalcraft.pathfinding.CoordinateTargetValidatable;
import ca.recrown.islandsurvivalcraft.pathfinding.CoordinateValidatable;
import ca.recrown.islandsurvivalcraft.pathfinding.DepthFirstSearch;
import ca.recrown.islandsurvivalcraft.world.BiomeSelector;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMapper;

//Note: technically, the validators have to be run on land, and so, some condition checks may not be nessecary.
public class BiomePerIslandGenerator implements IslandBiomeGenerator {
    private volatile boolean initialized;
    private final TemperatureMapGenerator temperatureMapGenerator;
    private final Cache<Point2, Biome[][]> chunkBiomesCache;
    private final Cache<Point2, Boolean> chunkGenStatusCache;
    private volatile IslandWorldMapper worldIslandMap;
    private volatile BiomeSelector biomeSelector;
    private volatile World world;
    private final DepthFirstSearch freshCachePropagator;
    private final DepthFirstSearch existenceChecker;
    private final FreshCachePropagationInfo freshCachePropInfo;
    private final PreviousGenerationInfo existenceInfo;
    private volatile Point2 currChunkCoords;
    private volatile Biome[][] localChunkCache;
    
    float temperature;

    public BiomePerIslandGenerator() {
        this.temperatureMapGenerator = new TemperatureMapGenerator();
        chunkBiomesCache = new Cache<>(4096);
        chunkGenStatusCache = new Cache<>(4096);
        freshCachePropInfo = new FreshCachePropagationInfo();
        freshCachePropagator = new DepthFirstSearch(freshCachePropInfo);
        existenceInfo = new PreviousGenerationInfo();
        existenceChecker = new DepthFirstSearch(existenceInfo);
        localChunkCache = new Biome[16][16];
    }
    
    @Override
    public void initialize(final World world, final IslandWorldMapper mapGenerator, final BiomeSelector biomeSelector) {
        if (initialized) throw new IllegalStateException("Biome generator already initialized.");
        initialized = true;
        this.world = world;
        this.worldIslandMap = mapGenerator;
        this.biomeSelector = biomeSelector;
        this.temperatureMapGenerator.setSeed(world.getSeed());
    }

    @Override
    public IslandBiomeGenerator getInstance() {
        return new BiomePerIslandGenerator();
    }

    @Override
    public Biome GenerateBiome(int chunkX, int chunkZ, int localX, int localZ) {
        if (currChunkCoords == null || chunkX != currChunkCoords.x || chunkZ != currChunkCoords.y) {
            currChunkCoords = new Point2(chunkX, chunkZ);
            localChunkCache = new Biome[16][16];
        }
        int worldX = Utilities.addMagnitude(16 * chunkX, localX);
        int worldZ = Utilities.addMagnitude(16 * chunkZ, localZ);

        Biome cachedBiome = getBiome(worldX, worldZ);
        if (cachedBiome != null) return cachedBiome;
        
        temperature = temperatureMapGenerator.getTemperature(worldX, worldZ);
        if (!worldIslandMap.isIsland(worldX, worldZ)) {
            return biomeSelector.getOceanBiome(temperature);
        }
        
        freshCachePropInfo.clear();
        existenceChecker.setStartPosition(worldX, worldZ);
        if (!existenceChecker.findTarget(existenceInfo)) {
            if (freshCachePropInfo.main == null) freshCachePropInfo.main = biomeSelector.getLandBiome(temperature);
            if (freshCachePropInfo.shore == null) freshCachePropInfo.shore = biomeSelector.getShoreBiome(freshCachePropInfo.main, temperature);
            if (freshCachePropInfo.shallow == null) freshCachePropInfo.shallow = biomeSelector.getOceanBiome(temperature);            
        }
        freshCachePropagator.setStartPosition(worldX, worldZ);
        freshCachePropagator.findTarget(freshCachePropInfo);
        return getBiome(worldX, worldZ);
    }

    private Biome getBiome(int worldX, int worldZ) {
        int localX = Math.abs(worldX % 16);
        int localZ = Math.abs(worldZ % 16);
        Point2 chunkCoords = new Point2(worldX / 16, worldZ / 16);

        if (chunkCoords.fastEquals(this.currChunkCoords) && localChunkCache[localX][localZ] != null) {
            return localChunkCache[localX][localZ];
        }

        Biome[][] biomes = chunkBiomesCache.getValue(chunkCoords);

        if (biomes != null) {
            Biome biome = biomes[localX][localZ];
            if (biome != null) return biome;
        }

        Boolean chunkGenStat = chunkGenStatusCache.getValue(chunkCoords);
        if (chunkGenStat == null) {
            chunkGenStat = world.isChunkGenerated(chunkCoords.x, chunkCoords.y);
            chunkGenStatusCache.setValue(chunkCoords, chunkGenStat);
        }
        
        if (chunkGenStat) {
            if (biomes == null) biomes = new Biome[16][16];
            biomes[localX][localZ] = world.getBiome(worldX, 0, worldZ);
            chunkBiomesCache.setValue(chunkCoords, biomes);
            return biomes[localX][localZ];
        }
        return null;
    }

    private void setCacheBiome(int worldX, int worldZ, Biome biome) {
        int localX = Math.abs(worldX % 16);
        int localZ = Math.abs(worldZ % 16);
        Point2 chunkCoords = new Point2(worldX / 16, worldZ / 16);

        if (chunkCoords.fastEquals(this.currChunkCoords)) localChunkCache[localX][localZ] = biome;

        Biome[][] chunkBiomes = chunkBiomesCache.getValue(chunkCoords);
        if (chunkBiomes == null) chunkBiomes = new Biome[16][16];
        chunkBiomes[localX][localZ] = biome;
        chunkBiomesCache.setValue(chunkCoords, chunkBiomes);
    }

    private class FreshCachePropagationInfo implements CoordinateTargetValidatable, CoordinateValidatable {
        public Biome shore;
        public Biome main;
        public Biome shallow;

        @Override
        public boolean isCoordinateTarget(int x, int y) {
            if (worldIslandMap.isLand(x, y)) {
                if (worldIslandMap.isShore(x, y)) {
                    setCacheBiome(x, y, shore);
                } else {
                    setCacheBiome(x, y, main);
                }
            } else {
                setCacheBiome(x, y, shallow);
            }
            return false;
        }
        
        @Override
        public boolean validate(int x, int y) {
            Point2 chunkCoords = new Point2(x / 16, y / 16);
            return chunkCoords.fastEquals(currChunkCoords) && worldIslandMap.isIsland(x, y);
        }

        public boolean allBiomesAcquired() {
            return shore != null && main != null && shallow != null;
        }

        public void clear() {
            main = null;
            shore = null;
            shallow = null;
        }
    }

    private class PreviousGenerationInfo implements CoordinateTargetValidatable, CoordinateValidatable {

        @Override
        public boolean validate(int x, int y) {
            return worldIslandMap.isIsland(x, y);
        }

        @Override
        public boolean isCoordinateTarget(int x, int y) {
            if (freshCachePropInfo.main == null && worldIslandMap.isLand(x, y) && !worldIslandMap.isShore(x, y)) {
                freshCachePropInfo.main = getBiome(x, y);
            }
            if (freshCachePropInfo.shore == null && worldIslandMap.isShore(x, y)) {
                freshCachePropInfo.shore = getBiome(x, y);
            }
            if (freshCachePropInfo.shallow == null && worldIslandMap.isShallowPortion(x, y)) {
                freshCachePropInfo.shallow = getBiome(x, y);
            }
            return freshCachePropInfo.allBiomesAcquired();
        }

    }
}